#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------
# Copyright 2019-2020 Airinnova AB and the FramAT authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------

# Author: Aaron Dettmann

"""
Plotting
"""

from commonlibs.math.vectors import unit_vector
from mpl_toolkits.mplot3d import Axes3D
import matplotlib.cm as cm
import matplotlib.pyplot as plt
import numpy as np

from ._element import GlobalSystem


class C:
    BC = 'black'
    BC_TEXT = 'white'
    CONC_FORCE = 'steelblue'
    CONC_MOMENT = 'purple'
    DEFAULT = 'grey'
    DEFORMED = 'red'
    DIST_FORCE = 'steelblue'
    DIST_MOMENT = 'purple'
    FREENODE_FORCE = 'black'
    FREENODE_MOMENT = 'grey'
    FREENODE_POINT = 'navy'
    GLOBAL_SYS = 'blue'
    LOCAL_SYS = 'green'
    MASS = 'maroon'
    MASS_LOAD = 'maroon'
    UNDEFORMED = 'blue'


def plot_all(m):
    """Create all plots defined in the model object"""

    mpp = m.get('post_proc', None)
    if not mpp:
        return
    if not mpp.get('plot', ()):
        return

    abm = m.results.get('mesh').get('abm')

    for plot_num, _ in enumerate(m.get('post_proc').iter('plot')):
        ax = init_3D_plot(*abm.get_lims())
        add_deformed_undeformed(m, ax, plot_num)

    plt.show()
    plt.close('all')


def init_3D_plot(x_lims, y_lims, z_lims):
    """
    Inititalize the 3D plot

    Args:
        :x_lims: (tuple) min and max x-value
        :y_lims: (tuple) min and max y-value
        :z_lims: (tuple) min and max z-value
    """

    fig = plt.figure(figsize=(10, 10))
    ax = fig.gca(projection='3d')

    # Avoid setting same min and max value by adding diff
    diff = (-1e-6, 1e-6)
    ax.set_xlim(*(x+d for x, d in zip(x_lims, diff)))
    ax.set_ylim(*(y+d for y, d in zip(y_lims, diff)))
    ax.set_zlim(*(z+d for z, d in zip(z_lims, diff)))

    set_equal_aspect_3D(ax)
    ax.set_xlabel('X')
    ax.set_ylabel('Y')
    ax.set_zlabel('Z')

    return ax


def set_equal_aspect_3D(ax):
    """
    Set aspect ratio of plot correctly

    Args:
        :ax: (obj) axis object
    """

    # See https://stackoverflow.com/a/19248731
    # ax.set_aspect('equal') --> raises a NotImplementedError
    # See https://github.com/matplotlib/matplotlib/issues/1077/

    extents = np.array([getattr(ax, 'get_{}lim'.format(dim))() for dim in 'xyz'])
    sz = extents[:, 1] - extents[:, 0]
    centers = np.mean(extents, axis=1)
    maxsize = max(abs(sz))
    r = maxsize/2
    for ctr, dim in zip(centers, 'xyz'):
        getattr(ax, 'set_{}lim'.format(dim))(ctr - r, ctr + r)


def args_plot(m, color, marker=None):
    args = {
        'linewidth': m.get('post_proc').get('plot_settings', {}).get('linewidth', 2),
        'markersize': m.get('post_proc').get('plot_settings', {}).get('markersize', 5),
        'color': color,
    }
    if marker is not None:
        args['marker'] = marker
    return args


def args_scatter(m, color, marker=None):
    args = {
        'linewidth': m.get('post_proc').get('plot_settings', {}).get('linewidth', 2),
        'color': color,
    }
    return args


def args_text(m, color):
    args = {
        'fontsize': m.get('post_proc').get('plot_settings', {}).get('fontsize', 10),
        'color': color,
        'bbox': dict(facecolor='orange', alpha=0.5),
        'horizontalalignment': 'center',
        'verticalalignment': 'bottom',
    }
    return args


def add_deformed_undeformed(m, ax, plot_num):
    to_show = m.get('post_proc').get('plot')[plot_num]
    abm = m.results.get('mesh').get('abm')
    marker = 'o' if 'nodes' in to_show else None

    for beam_id in abm.beams.keys():
        xyz = abm.get_sup_points(beam_id)
        x, y, z = xyz[:, 0], xyz[:, 1], xyz[:, 2]
        ax.plot(x, y, z, **args_plot(m, C.UNDEFORMED, marker=marker))

        # ==================
        d = m.results.get('beam')[0].get('deformation')
        xd = x + d['ux']
        yd = y + d['uy']
        zd = z + d['uz']
        ax.plot(xd, yd, zd, **args_plot(m, C.DEFORMED, marker=marker))
        # ==================


    # if 'node_uids' in to_show:
        # for uid, coord in abm.named_nodes.items():
        #     ax.text(*coord, uid, **args_text(m, color=C.BC))


# def plot_all(frame, plots, filestructure):
#     """
#     Wrapper function which creates all plot defined in 'plots'

#     Args:
#         :frame: frame object
#         :plots: input (list with plot settings as dict)
#     """

#     n_plots = len(plots)

#     for i, settings in enumerate(plots, start=1):
#         settings['plot_number'] = i

#         if settings.get('show', True) or settings.get('set_savefig', False):
#             logger.info(f"Creating plot {i}/{n_plots}")

#             if settings.get('matrix_plots', False):
#                 create_matrix_plot(frame, settings)
#             else:
#                 create_3D_plot(frame, settings, filestructure)
#         else:
#             logger.info(f"Skipping plot {i}/{n_plots}")


# def get_all_plot_settings(ps, frame):
#     """
#     Return a dictionary with all plot setting

#     Args:
#         :plot_settings: dict with settings
#         :frame: frame object

#     Returns:
#         :plot_settings: completed dict (defaults added if no value provided)
#     """

#     for default in PLOT3D_DEFAULTS:
#         setting, value = default
#         ps[setting] = ps.get(setting, value)

#     # Other settings
#     ps['marker'] = '.' if ps['plot_nodes'] else None
#     ps['frame'] = frame
#     ps['D'] = frame.deformation.U

#     return ps

    # if ps['plot_global_axes']:
    #     origin_glob = GlobalSystem.UnitVectors.Origin
    #     x_glob = GlobalSystem.UnitVectors.X
    #     y_glob = GlobalSystem.UnitVectors.Y
    #     z_glob = GlobalSystem.UnitVectors.Z
    #     ax = _coordinate_system(ax, ps, origin_glob,
    #                             (x_glob, y_glob, z_glob), ('X', 'Y', 'Z'),
    #                             color=COLOR_GLOBAL_SYS)

    # if ps['plot_local_axes']:
    #     for beamline in frame.beamlines:
    #         for element in beamline.elements:
    #             mid_point = element.mid_point

    #             if ps['plot_local_axes']:
    #                 axes = (element.x_elem, element.y_elem, element.z_elem)
    #                 axes_names = ('x', 'y', 'z')
    #                 scale = 0.4*element.length
    #                 ax = _coordinate_system(ax, ps, mid_point, axes, axes_names, COLOR_LOCAL_SYS, scale)

    # if ps['plot_accel_vector']:
    #     ax = _accel_vector(ax, ps)

    # if ps['plot_undeformed']:
    #     ax = _undeformed(ax, ps)

    # if ps['plot_named_nodes']:
    #     ax = _node_uids(ax, ps)

    # if ps['plot_beamline_uids']:
    #     ax = _beamline_uids(ax, ps)

    # if ps['plot_free_nodes'] or ps['plot_free_node_vectors']:
    #     ax = _free_nodes(ax, ps)

    # if ps['plot_bc'] or ps['plot_bc_labels']:
    #     ax = _boundary_conditions(ax, ps)

    # if ps['plot_inertia_symbols']:
    #     ax = _inertia_symbols(ax, ps)

    # if ps['plot_inertia_loads'] or ps['plot_inertia_load_labels']:
    #     ax = _inertia_loads(ax, ps)

    # if ps['plot_loads']:
    #     ax = _loads(ax, ps)

    # if ps['plot_deformed']:
    #     ax = _deformed(ax, ps)

    # if ps['plot_CG_total'] or ps['plot_CG_beams']:
    #     ax = _centre_of_mass(ax, ps)

    # plt.tight_layout()


# def _coordinate_system(plot, ps, origin, axes, axes_names, color, scale=1):
#     """
#     Add a coordinate system to a plot

#     Args:
#         :plot: plot
#         :ps: plot settings
#         :origin: origin of coordinate system
#         :axes: tuple with axes (x, y, z)
#         :axes_names: tuple with names of the axes (e.g. ('x', 'y', 'z'))
#         :scale: scalar to size the length of the axis vectors

#     Returns:
#         :plot: modified plot object
#     """

#     axes = [scale*np.array(axis) for axis in axes]
#     x_axis, y_axis, z_axis = axes

#     for axis, axis_name in zip(axes, axes_names):
#         x, y, z = origin
#         u, v, w = axis
#         plot.scatter(x, y, z, color=color, linewidth=ps['set_linewidth'])
#         plot.quiver(x, y, z, u, v, w, length=1, color=color, linewidth=ps['set_vectorwidth'])
#         plot.text(x+u, y+v, z+w, axis_name, color=color, fontsize=ps['set_fontsize'])

#     # Plot xy-plane
#     p1 = np.array(origin)
#     p2 = np.array(origin + y_axis)
#     p3 = np.array(origin + z_axis)
#     p4 = np.array(origin + y_axis + z_axis)
#     points = np.array([p1, p2, p3, p4])
#     xx = points[:, 0].reshape(2, 2)
#     yy = points[:, 1].reshape(2, 2)
#     z = points[:, 2].reshape(2, 2)
#     plot.plot_surface(xx, yy, z, alpha=0.4, color=color)

#     return plot


# def _undeformed(plot, ps):
#     """
#     Add the undeformed frame geometry to a plot

#     Args:
#         :plot: plot
#         :ps: plot settings

#     Returns:
#         :plot: modified plot object
#     """

#     frame = ps['frame']

#     for beamline in frame.beamlines:
#         for element in beamline.elements:

#             # Plot elements
#             node_points = np.array([element.node1.coord, element.node2.coord])
#             x = node_points[:, 0]
#             y = node_points[:, 1]
#             z = node_points[:, 2]

#             plot.plot(x, y, z, linewidth=ps['set_linewidth'],
#                       marker=ps['marker'], markersize=ps['set_markersize'], color=COLOR_UNDEFORMED)

#     return plot


# def _deformed(plot, ps):
#     """
#     Add the deformed frame geometry to a plot

#     Args:
#         :plot: plot
#         :ps: plot settings

#     Returns:
#         :plot: modified plot object
#     """

#     frame = ps['frame']
#     D = ps['D']
#     u_factor = ps['set_amplify_displacement']
#     t_factor = ps['set_amplify_rotation']

#     i = 0
#     for beamline in frame.beamlines:
#         x_def = []
#         y_def = []
#         z_def = []

#         node_y_axis = []
#         theta_x = []
#         theta_y = []
#         theta_z = []

#         for node in frame.finder.nodes.by_beamline_uid[beamline.uid]:
#             x, y, z = node.coord

#             x_def.append(x + u_factor*D[6*i, 0])
#             y_def.append(y + u_factor*D[6*i+1, 0])
#             z_def.append(z + u_factor*D[6*i+2, 0])

#             if ps['plot_rotations']:
#                 node_y_axis.append(node.parent_element.y_elem)
#                 theta_x.append(t_factor*D[6*i+3, 0])
#                 theta_y.append(t_factor*D[6*i+4, 0])
#                 theta_z.append(t_factor*D[6*i+5, 0])

#             i += 1

#         # PLOT TRANSLATION
#         plot.plot(x_def, y_def, z_def, linewidth=ps['set_linewidth'],
#                   marker=ps['marker'], markersize=ps['set_markersize'], color=COLOR_DEFORMED)

#         # PLOT ROTATION
#         # TODO: there must be a better way
#         if ps['plot_rotations']:
#             j = 0
#             for y_ax, tx, ty, tz in zip(node_y_axis, theta_x, theta_y, theta_z):
#                 # Using local y-axis to indicate rotation
#                 y_ax = rotate_vector_around_axis(y_ax, GlobalSystem.UnitVectors.X, tx)
#                 y_ax = rotate_vector_around_axis(y_ax, GlobalSystem.UnitVectors.Y, ty)
#                 y_ax = rotate_vector_around_axis(y_ax, GlobalSystem.UnitVectors.Z, tz)

#                 p1 = np.array([
#                               x_def[j] - y_ax[0],
#                               y_def[j] - y_ax[1],
#                               z_def[j] - y_ax[2]])

#                 p2 = np.array([
#                               x_def[j] + y_ax[0],
#                               y_def[j] + y_ax[1],
#                               z_def[j] + y_ax[2]])

#                 node_points = np.array([p1, p2])
#                 x = node_points[:, 0]
#                 y = node_points[:, 1]
#                 z = node_points[:, 2]

#                 plot.plot(x, y, z, marker=ps['marker'], linewidth=0.2*ps['set_linewidth'], color=COLOR_DEFORMED)
#                 j += 1

#     # Indicate rigid connection in deformed state
#     bc = frame.boundary_conditions.boundary_conditions

#     if ps['plot_bc'] and 'connect' in bc:
#         for connection in bc['connect']:
#             node1_name = connection['node1']
#             node2_name = connection['node2']
#             node1_coord = frame.finder.nodes.by_uid[node1_name].coord
#             node2_coord = frame.finder.nodes.by_uid[node2_name].coord
#             node1_num = frame.finder.nodes.by_uid[node1_name].num
#             node2_num = frame.finder.nodes.by_uid[node2_name].num
#             def_node1 = D[6*node1_num:6*node1_num+3, 0]*ps['set_amplify_displacement']
#             def_node2 = D[6*node2_num:6*node2_num+3, 0]*ps['set_amplify_displacement']

#             node1_coord = node1_coord + def_node1
#             node2_coord = node2_coord + def_node2

#             node_points = np.array([node1_coord, node2_coord])
#             x = node_points[:, 0]
#             y = node_points[:, 1]
#             z = node_points[:, 2]

#             plot.plot(x, y, z, ':', linewidth=0.5*ps['set_linewidth'], color=COLOR_BC)

#     return plot


# def _boundary_conditions(plot, ps):
#     """
#     Add text label indicating the imposed boundary conditions

#     Args:
#         :plot: plot
#         :ps: plot settings

#     Returns:
#         :plot: modified plot object
#     """

#     frame = ps['frame']
#     bc = frame.boundary_conditions.boundary_conditions

#     if 'fix' in bc:
#         for entry in bc['fix']:
#             node_name, constraints = entry['node'], entry['fix']

#             node = frame.finder.nodes.by_uid[node_name]
#             coord = node.coord
#             bc_id = frame.boundary_conditions.get_identifier(constraints)
#             bc_id = f"f{bc_id}"

#             if ps['plot_bc']:
#                 plot.scatter(*coord, marker='D', color=COLOR_BC, s=20)

#             if ps['plot_bc_labels']:
#                 plot.text(*coord, bc_id, bbox=dict(facecolor='grey', alpha=0.9),
#                           horizontalalignment='center', verticalalignment='top',
#                           color=COLOR_BC_TEXT, fontsize=ps['set_fontsize'])

#     if 'connect' in bc:
#         for connection in bc['connect']:
#             node1_name = connection['node1']
#             node2_name = connection['node2']
#             node1_coord = frame.finder.nodes.by_uid[node1_name].coord
#             node2_coord = frame.finder.nodes.by_uid[node2_name].coord
#             bc_id = frame.boundary_conditions.get_identifier(connection['fix'])
#             bc_id = f"c{bc_id}"

#             node_points = np.array([node1_coord, node2_coord])
#             x = node_points[:, 0]
#             y = node_points[:, 1]
#             z = node_points[:, 2]

#             if ps['plot_bc']:
#                 plot.plot(x, y, z, '-', linewidth=ps['set_linewidth'], color=COLOR_BC)

#             if ps['plot_bc_labels']:
#                 mid_point = (node_points[0, :] + node_points[1, :])/2
#                 plot.text(*mid_point, bc_id,
#                           bbox=dict(facecolor='grey', alpha=0.9),
#                           horizontalalignment='center', verticalalignment='top',
#                           color=COLOR_BC_TEXT, fontsize=ps['set_fontsize'])

#     return plot


# def _node_uids(plot, ps):
#     """
#     Add text label indicating nodes with UIDs

#     Args:
#         :plot: plot
#         :ps: plot settings

#     Returns:
#         :plot: modified plot object
#     """

#     frame = ps['frame']

#     for node_uid, node in frame.finder.nodes.by_uid.items():
#         coord = node.coord
#         plot.text(*coord, node_uid,
#                   bbox=dict(facecolor='orange', alpha=0.5),
#                   horizontalalignment='center', verticalalignment='bottom',
#                   color=COLOR_BC, fontsize=ps['set_fontsize'])
#     return plot


# def _beamline_uids(plot, ps):
#     """
#     Add text label indicating UIDs of beamlines

#     Args:
#         :plot: plot
#         :ps: plot settings

#     Returns:
#         :plot: modified plot object
#     """

#     frame = ps['frame']

#     for beam_uid, beam in frame.finder.beamlines.by_uid.items():
#         coord = beam.interpolator(0.5)
#         plot.text(*coord, beam_uid,
#                   bbox=dict(facecolor='navy', alpha=0.5),
#                   horizontalalignment='center', verticalalignment='bottom',
#                   color='white', fontsize=ps['set_fontsize'])
#     return plot


# def _loads(plot, ps):
#     """
#     Add vectors and text labels indicating externally applied loads

#     Args:
#         :plot: plot
#         :ps: plot settings

#     Returns:
#         :plot: modified plot object
#     """

#     frame = ps['frame']

#     for beamline in frame.beamlines:
#         for element in beamline.elements:
#             # ----- Concentrated loads -----
#             f_c = element.concentrated_load_vector_glob
#             x1, y1, z1 = element.node1.coord
#             x2, y2, z2 = element.node2.coord

#             F1 = f_c[0:3, 0]
#             M1 = f_c[3:6, 0]
#             F2 = f_c[6:9, 0]
#             M2 = f_c[9:12, 0]

#             F1_mag = np.linalg.norm(F1)
#             M1_mag = np.linalg.norm(M1)
#             F2_mag = np.linalg.norm(F2)
#             M2_mag = np.linalg.norm(M2)

#             F1 = F1*ps['set_scale_conc_loads']
#             M1 = M1*ps['set_scale_conc_moments']
#             F2 = F2*ps['set_scale_conc_loads']
#             M2 = M2*ps['set_scale_conc_moments']

#             if np.linalg.norm(F1) > 0:
#                 plot.scatter(x1, y1, z1, color=COLOR_CONC_FORCE, linewidth=ps['set_vectorwidth'])
#                 plot.quiver(x1, y1, z1, F1[0], F1[1], F1[2],
#                             length=1, color=COLOR_CONC_FORCE, linewidth=ps['set_vectorwidth'])
#                 if ps['plot_load_labels']:
#                     plot.text(x1+F1[0], y1+F1[1], z1+F1[2], f"{F1_mag:.1e} N",
#                               bbox=dict(facecolor=COLOR_CONC_FORCE, alpha=0.5),
#                               color='black', fontsize=ps['set_fontsize'])

#             if np.linalg.norm(M1) > 0:
#                 plot.scatter(x1, y1, z1, color=COLOR_CONC_MOMENT, linewidth=ps['set_vectorwidth'])
#                 plot.quiver(x1, y1, z1, M1[0], M1[1], M1[2],
#                             length=1, color=COLOR_CONC_MOMENT, linewidth=ps['set_vectorwidth'])
#                 if ps['plot_load_labels']:
#                     plot.text(x1+M1[0], y1+M1[1], z1+M1[2], f"{M1_mag:.1e} Nm",
#                               bbox=dict(facecolor=COLOR_CONC_MOMENT, alpha=0.5),
#                               color='black', fontsize=ps['set_fontsize'])

#             if np.linalg.norm(F2) > 0:
#                 plot.quiver(x2, y2, z2, F2[0], F2[1], F2[2],
#                             length=1, color=COLOR_CONC_FORCE, linewidth=ps['set_vectorwidth'])
#                 plot.scatter(x2, y2, z2, color=COLOR_CONC_FORCE, linewidth=ps['set_vectorwidth'])
#                 if ps['plot_load_labels']:
#                     plot.text(x2+F2[0], y2+F2[1], z2+F2[2], f"{F2_mag:.1e} N",
#                               bbox=dict(facecolor=COLOR_CONC_FORCE, alpha=0.5),
#                               color='black', fontsize=ps['set_fontsize'])

#             if np.linalg.norm(M2) > 0:
#                 plot.quiver(x2, y2, z2, M2[0], M2[1], M2[2],
#                             length=1, color=COLOR_CONC_MOMENT, linewidth=ps['set_vectorwidth'])
#                 plot.scatter(x2, y2, z2, color=COLOR_CONC_MOMENT, linewidth=ps['set_vectorwidth'])
#                 if ps['plot_load_labels']:
#                     plot.text(x2+M2[0], y2+M2[1], z2+M2[2], f"{M2_mag:.1e} Nm",
#                               bbox=dict(facecolor=COLOR_CONC_MOMENT, alpha=0.5),
#                               color='black', fontsize=ps['set_fontsize'])

#             # ----- Distributed loads -----
#             f_d = element.distributed_load_vector_glob
#             x, y, z = element.mid_point

#             dist_force = f_d[0:3, 0] + f_d[6:9, 0]
#             dist_moment = f_d[3:6, 0] + f_d[9:12, 0]

#             dist_force_mag = np.linalg.norm(dist_force)
#             dist_moment_mag = np.linalg.norm(dist_moment)

#             dist_force = dist_force*ps['set_scale_dist_loads']
#             dist_moment = dist_moment*ps['set_scale_dist_loads']

#             if np.linalg.norm(dist_force) > 0:
#                 plot.scatter(x, y, z, color=COLOR_DIST_FORCE, linewidth=ps['set_vectorwidth'])
#                 plot.quiver(x, y, z, dist_force[0], dist_force[1], dist_force[2],
#                             length=1, color=COLOR_DIST_FORCE, linewidth=2*ps['set_vectorwidth'])
#                 if ps['plot_load_labels']:
#                     plot.text(x+dist_force[0], y+dist_force[1], z+dist_force[2],
#                               f"{dist_force_mag:.1e} N",
#                               bbox=dict(facecolor=COLOR_DIST_FORCE, alpha=0.5),
#                               color='black', fontsize=ps['set_fontsize'])

#             if np.linalg.norm(dist_moment) > 0:
#                 plot.scatter(x, y, z, color=COLOR_DIST_MOMENT, linewidth=ps['set_vectorwidth'])
#                 plot.quiver(x, y, z, dist_moment[0], dist_moment[1], dist_moment[2],
#                             linewidth=2*ps['set_vectorwidth'], length=1, color=COLOR_DIST_MOMENT)
#                 if ps['plot_load_labels']:
#                     plot.text(x+dist_moment[0], y+dist_moment[1], z+dist_moment[2],
#                               f"{dist_moment_mag:.1e} N",
#                               bbox=dict(facecolor=COLOR_DIST_MOMENT, alpha=0.5),
#                               color='black', fontsize=ps['set_fontsize'])

#     return plot


# def _inertia_symbols(plot, ps):
#     """
#     Add symbols indication masses

#     Args:
#         :plot: plot
#         :ps: plot settings

#     Returns:
#         :plot: modified plot object
#     """

#     frame = ps['frame']

#     for beamline in frame.beamlines:
#         for element in beamline.elements:
#             # ----- Point masses -----
#             m1 = element.point_properties['m1']
#             m2 = element.point_properties['m2']

#             if m1 != 0:
#                 coord = element.node1.coord
#                 plot.scatter(*coord, color=COLOR_MASS, marker='s', linewidth=ps['set_linewidth'])
#                 plot.text(*coord, f'{m1:.1e} kg',
#                           bbox=dict(facecolor='grey', alpha=0.9),
#                           horizontalalignment='left', verticalalignment='top',
#                           color=COLOR_BC_TEXT, fontsize=ps['set_fontsize'])

#             if m2 != 0:
#                 coord = element.node2.coord
#                 plot.scatter(*coord, color=COLOR_MASS, marker='s', linewidth=ps['set_linewidth'])
#                 plot.text(*coord, f'{m2:.1e} kg',
#                           bbox=dict(facecolor='grey', alpha=0.9),
#                           horizontalalignment='left', verticalalignment='top',
#                           color=COLOR_BC_TEXT, fontsize=ps['set_fontsize'])
#     return plot


# def _free_nodes(plot, ps):
#     """
#     Add free nodes and free node loads

#     Args:
#         :plot: plot
#         :ps: plot settings

#     Returns:
#         :plot: modified plot object
#     """

#     frame = ps['frame']

#     for beamline in frame.beamlines:
#         for f_node in beamline.free_nodes:
#             coord = f_node['coord']
#             load = f_node['load']
#             closest = f_node['closest_beam_node']

#             points = np.array([coord, closest.coord])
#             x = points[:, 0]
#             y = points[:, 1]
#             z = points[:, 2]

#             if ps['plot_free_nodes']:
#                 plot.scatter(*coord, color=COLOR_FREENODE_POINT, marker='.', linewidth=ps['set_linewidth'])
#                 plot.plot(x, y, z, color=COLOR_FREENODE_POINT, linewidth=0.5*ps['set_linewidth'])

#             if ps['plot_free_node_vectors']:
#                 plot.quiver(*coord, *load[0:3], length=ps['set_scale_free_node_force'],
#                             color=COLOR_FREENODE_FORCE, linewidth=0.5*ps['set_vectorwidth'])
#                 plot.quiver(*coord, *load[3:6], length=ps['set_scale_free_node_force'],
#                             color=COLOR_FREENODE_MOMENT, linewidth=0.5*ps['set_vectorwidth'])

#     return plot


# def _accel_vector(plot, ps):
#     """
#     Add a vector indicating the direction of (gravitational) acceleration

#     Args:
#         :plot: plot
#         :ps: plot settings

#     Returns:
#         :plot: modified plot object
#     """

#     frame = ps['frame']

#     accel_direction = frame.accel_direction
#     accel_magnitude = np.linalg.norm(accel_direction)
#     orig = 3*GlobalSystem.UnitVectors.Z
#     vector = unit_vector(accel_direction)

#     plot.scatter(*orig, color=COLOR_BC, linewidth=ps['set_linewidth'])
#     plot.quiver(*orig, *vector, length=1, color=COLOR_BC, linewidth=ps['set_vectorwidth'])
#     plot.text(*orig, f"{accel_magnitude:.2e} m/s²", color=COLOR_BC, fontsize=ps['set_fontsize'])

#     return plot


# def _inertia_loads(plot, ps):
#     """
#     Add vectors and text labels indicating inertial loads

#     Args:
#         :plot: plot
#         :ps: plot settings

#     Returns:
#         :plot: modified plot object
#     """

#     frame = ps['frame']
#     F_accel = frame.F_accel

#     for i, node in frame.finder.nodes.by_number.items():
#         load = F_accel[i*6:i*6+6]

#         F = load[0:3]
#         M = load[3:6]
#         F_mag = np.linalg.norm(F)
#         M_mag = np.linalg.norm(M)
#         F = F*ps['set_scale_inertia_loads']
#         M = M*ps['set_scale_inertia_loads']

#         if np.linalg.norm(F) > 0:
#             plot.scatter(*node.coord, color=COLOR_MASS_LOAD, linewidth=ps['set_vectorwidth'])
#             plot.quiver(*node.coord, *F, length=1, color=COLOR_MASS_LOAD, linewidth=2*ps['set_vectorwidth'])

#             if ps['plot_inertia_load_labels']:
#                 plot.text(*(node.coord+F[0][:]), f"{F_mag:.1e} N",
#                           bbox=dict(facecolor=COLOR_MASS_LOAD, alpha=0.5),
#                           color='black', fontsize=ps['set_fontsize'])

#         if np.linalg.norm(M) > 0:
#             plot.scatter(*node.coord, color=COLOR_MASS_LOAD, linewidth=ps['set_vectorwidth'])
#             plot.quiver(*node.coord, *M, length=1, color=COLOR_MASS_LOAD, linewidth=2*ps['set_vectorwidth'])

#             if ps['plot_inertia_load_labels']:
#                 plot.text(*(node.coord+M[0][:]), f"{M_mag:.1e} N*m",
#                           bbox=dict(facecolor=COLOR_MASS_LOAD, alpha=0.5),
#                           color='black', fontsize=ps['set_fontsize'])

#     return plot


# def _centre_of_mass(plot, ps):
#     """
#     Add marker indicating the centre of mass of the entire frame of beams

#     Args:
#         :plot: plot
#         :ps: plot settings

#     Returns:
#         :plot: modified plot object
#     """

#     frame = ps['frame']
#     mass_breakdown = frame.get_mass_breakdown()

#     if ps['plot_CG_total']:
#         coord_CG = mass_breakdown['total_CG']
#         total_mass = mass_breakdown['total_mass']

#         plot.scatter(*coord_CG, color='black', marker='X', s=ps['set_markersize']*10)
#         plot.text(*coord_CG, f"  CG: {total_mass:.1e} kg",
#                   bbox=dict(facecolor='grey', alpha=0.5),
#                   color='black', fontsize=ps['set_fontsize'])

#     if ps['plot_CG_beams']:
#         for beamline_uid, beamline_entry in mass_breakdown['beams'].items():
#             coord_CG = beamline_entry['CG']
#             total_mass = beamline_entry['mass']

#             plot.scatter(*coord_CG, color='black', marker='X', s=ps['set_markersize']*10)
#             plot.text(*coord_CG, f"  CG {beamline_uid}: {total_mass:.1e} kg",
#                       bbox=dict(facecolor='lightgrey', alpha=0.5),
#                       color='black', fontsize=ps['set_fontsize'])

#     return plot


# def create_matrix_plot(frame, plot_settings):
#     """
#     Create a matrix plot

#     Args:
#         :frame: frame object
#         :plot_settings: dict with settings
#     """

#     colormap = cm.get_cmap('magma')

#     matrix_name_list = plot_settings['matrix_plots']['matrices']
#     for i, matrix_name in enumerate(matrix_name_list, start=1):
#         try:
#             matrix = getattr(frame, matrix_name)
#         except AttributeError:
#             raise AttributeError(f"Could not find matrix '{matrix_name}'") from None

#         logger.info(f"--> Creating matrix plot {i}/{len(matrix_name_list)}")

#         figure = plt.figure(figsize=(9, 9))
#         axes = figure.add_subplot(111)
#         axes.set_aspect('equal')
#         axes.matshow(matrix, cmap=colormap)
#         plt.title(f"Matrix {matrix_name}")
#         plt.show()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------
# Copyright 2019-2020 Airinnova AB and the FramAT authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------

# Author: Aaron Dettmann

"""
Run the model
"""

from . import MODULE_NAME
from . import _assembly as ass
from . import _meshing as mesh
from . import _plot as plot
from . import _solve as sol
from .__version__ import __version__
from ._log import logger


def run_model(m):
    """
    Run the complete model analysis

    Args:
        :m: instance of model
    """

    logger.info(f"===== {MODULE_NAME} {__version__} =====")

    # Instantiate result storage for each beam
    for mbeam in m.iter('beam'):
        m.results.add_feature('beam')

    # ----- MESHING -----
    logger.info("Meshing...")
    mesh.create_mesh(m)

    # ----- ASSEMBLING SYSTEM MATRICES -----
    logger.info("Assembling matrices...")
    ass.create_system_matrices(m)
    ass.create_bc_matrices(m)

    # ----- SOLVING -----
    logger.info("Solving...")
    sol.solve(m)

    # ----- POST-PROCESSING -----
    logger.info("Post-processing...")
    plot.plot_all(m)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------
# Copyright 2019-2020 Airinnova AB and the FramAT authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------

# Author: Aaron Dettmann

"""
Beam element definition
"""

from collections import defaultdict
import logging

import numpy as np
from commonlibs.math.vectors import unit_vector, direction_cosine, vector_rejection

logger = logging.getLogger(__name__)


class GlobalSystem:
    Origin = np.array([0, 0, 0])
    X = np.array([1, 0, 0])
    Y = np.array([0, 1, 0])
    Z = np.array([0, 0, 1])


G = GlobalSystem


class Element:
    """
    Euler-Bernoulli element with 6 degrees of freedom

    Nomenclature

    Material properties:
        :E: Young's modulus [N/m²]
        :G: Shear modulus [N/m²]
        :rho: Material density [kg/m³]

    Cross section properties:
        :A: Cross section area [m²]
        :Iy: Second moment of area about element local y-axis [m⁴]
        :Iz: Second moment of area about element local z-axis [m⁴]
        :J: Torsional constant [m⁴]

    Inertia properties:
        :m{1,2}: Point mass attached to node 1 or 2 [kg]

    Loads:
        F{x,y,z}{1,2}: Point load in x,y,z-directions applied to node 1 or 2 [N]
        M{x,y,z}{1,2}: Point moment in x,y,z-directions applied to node 1 or 2 [N]
        q{x,y,z}: Line force in x,y,z-directions applied to length of element [N/m]
        m{x,y,z}: Line moment in x,y,z-directions applied to length of element [N*m/m]

    Deformation:
        u{x,y,z}{1,2}: Displacement in x,y,z-direction (in global system) [m]
        t{x,y,z}{1,2}: Rotation in x,y,z-direction (in global system) [rad]
    """

    DOF_PER_NODE = 6

    MATERIAL_PROPS = ('E', 'G', 'rho')
    CROSS_SECTION_PROPS = ('A', 'Iy', 'Iz', 'J')
    PROP_TYPES = MATERIAL_PROPS + CROSS_SECTION_PROPS

    POINT_PROPS = ('m1', 'm2')

    DEFORM_TYPES = ('ux', 'uy', 'uz', 'tx', 'ty', 'tz')
    DIST_LOAD_TYPES = ('qx', 'qy', 'qz', 'mx', 'my', 'mz')
    CONC_LOAD_BASE_TYPES = ('Fx', 'Fy', 'Fz', 'Mx', 'My', 'Mz')
    CONC_LOAD_TYPES = ('Fx1', 'Fy1', 'Fz1', 'Mx1', 'My1', 'Mz1',
                       'Fx2', 'Fy2', 'Fz2', 'Mx2', 'My2', 'Mz2')

    def __init__(self, p1, p2, up):
        """
        Beam finite element with 6 dof per node

        TODO
        """

        # ===== Geometry =====
        # Node points
        self.p1 = p1
        self.p2 = p2

        # Vectors of the local coordinate system
        self.x_elem = unit_vector(self.p2.coord - self.p1.coord)
        self.y_elem, self.z_elem = get_local_system_from_up(self.x_elem, up)

        # Additional geometric properties
        self.mid_point = (self.p1.coord + self.p2.coord)/2
        self.mid_xsi = (self.p1.rel_coord + self.p2.rel_coord)/2
        self.length = np.linalg.norm(self.p2.coord - self.p1.coord)

        # ===== Material and cross section properties =====
        self.properties = defaultdict(lambda: None)

        # ===== Load vector in the global system =====
        self.load_vector_glob = np.zeros((12, 1))
        self.mass_matrix_glob = np.zeros((12, 12))
        # self.stiffness_matrix_glob = np.zeros((12, 1))

        # ===== Transformation matrix =====
        lx = direction_cosine(self.x_elem, G.X)
        ly = direction_cosine(self.y_elem, G.X)
        lz = direction_cosine(self.z_elem, G.X)
        mx = direction_cosine(self.x_elem, G.Y)
        my = direction_cosine(self.y_elem, G.Y)
        mz = direction_cosine(self.z_elem, G.Y)
        nx = direction_cosine(self.x_elem, G.Z)
        ny = direction_cosine(self.y_elem, G.Z)
        nz = direction_cosine(self.z_elem, G.Z)

        T3 = np.array([[lx, mx, nx], [ly, my, ny], [lz, mz, nz]])
        self.T = np.zeros((12, 12))
        self.T[0:3, 0:3] = self.T[3:6, 3:6] = self.T[6:9, 6:9] = self.T[9:12, 9:12] = T3

    @classmethod
    def from_abstract_element(cls, a):
        """Create an new element from an abstract beam element"""

        new = cls(a.p1, a.p2, a.get('up'))

        for prop_type in new.PROP_TYPES:
            new.properties[prop_type] = a.get(prop_type)

        for pl in a.iter('point_load'):
            new.add_point_load(pl['load'], pl['node'], pl.get('loc_system', False))

        return new

    def add_point_load(self, load, node_num, loc_system):
        """
        Add a point load to the element node 1 or 2

        Args:
            :loads: dict with concentrated loads
            :loc_node_num: node to which loads are to be applied (1, 2)
            :loc_system: flag, which, if True, makes loads be treated according to local coordinate system
        """

        load = np.array(load).reshape((6, 1))
        load_contrib = np.zeros((12, 1))

        if node_num == 1:
            load_contrib[:6, ] = load
        else:
            load_contrib[6:, ] = load

        if loc_system:
            load_contrib = self.T @ load_contrib

        self.load_vector_glob += load_contrib

    #################################################################
    #################################################################
    #################################################################

    def shape_function_matrix(self, xi):
        """
        Return the shape function (basis function) matrix of shape (6, 12)

        Notes:
            * Rows 1 to 6 correspond to deformations 'ux', 'uy', 'uz',
              'tx', 'ty', 'tz' in this order

        Args:
            :xi: relative element coordinate [0, 1]

        Returns:
            :N: shape function matrix
        """

        if not (0 <= xi <= 1):
            raise ValueError("xi must be in range [0, 1]")

        L = self.length

        # Notes:
        # * At xi = 0: N1 = N3 = M1 = M5 = 1 (rest is 0)
        # * At xi = 1: N2 = N4 = M2 = M6 = 1 (rest is 0)

        N1 = 1 - xi
        N2 = xi
        N3 = 1 - 3*xi**2 + 2*xi**3
        N4 = 3*xi**2 - 2*xi**3
        N5 = L*(xi - 2*xi**2 + xi**3)
        N6 = L*(-xi**2 + xi**3)

        M1 = 1 - xi
        M2 = xi
        M3 = -(6/L)*(xi - xi**2)
        M4 = (6/L)*(xi - xi**2)
        M5 = 1 - 4*xi + 3*xi**2
        M6 = -2*xi + 3*xi**2

        N = np.zeros((6, 12))

        N[0, 0] = N1
        N[0, 6] = N2

        N[1, 1] = N3
        N[1, 5] = N5
        N[1, 7] = N4
        N[1, 11] = N6

        N[2, 2] = N3
        N[2, 4] = -N5
        N[2, 8] = N4
        N[2, 10] = -N6

        N[3, 3] = M1
        N[3, 9] = M2

        N[4, 2] = M3
        N[4, 4] = M5
        N[4, 8] = M4
        N[4, 10] = M6

        N[5, 1] = -M3
        N[5, 5] = M5
        N[5, 7] = -M4
        N[5, 11] = M6

        return N

    @property
    def stiffness_matrix_local(self):
        """Element stiffness matrix (as formulated in local system)"""

        E = self.properties['E']
        G = self.properties['G']
        A = self.properties['A']
        Iy = self.properties['Iy']
        Iz = self.properties['Iz']
        J = self.properties['J']

        EA = E*A
        EIy = E*Iy
        EIz = E*Iz
        GJ = G*J
        L = self.length
        L2 = L**2
        L3 = L**3

        k_elem = np.zeros((12, 12))

        k_elem[0, 0] = EA/L
        k_elem[0, 6] = -EA/L

        k_elem[1, 1] = 12*EIz/L3
        k_elem[1, 5] = 6*EIz/L2
        k_elem[1, 7] = -12*EIz/L3
        k_elem[1, 11] = 6*EIz/L2

        k_elem[2, 2] = 12*EIy/L3
        k_elem[2, 4] = -6*EIy/L2
        k_elem[2, 8] = -12*EIy/L3
        k_elem[2, 10] = -6*EIy/L2

        k_elem[3, 3] = GJ/L
        k_elem[3, 9] = -GJ/L

        k_elem[4, 4] = 4*EIy/L
        k_elem[4, 8] = 6*EIy/L2
        k_elem[4, 10] = 2*EIy/L

        k_elem[5, 5] = 4*EIz/L
        k_elem[5, 7] = -6*EIz/L2
        k_elem[5, 11] = 2*EIz/L

        k_elem[6, 6] = EA/L

        k_elem[7, 7] = 12*EIz/L3
        k_elem[7, 11] = -6*EIz/L2

        k_elem[8, 8] = 12*EIy/L3
        k_elem[8, 10] = 6*EIy/L2

        k_elem[9, 9] = GJ/L

        k_elem[10, 10] = 4*EIy/L

        k_elem[11, 11] = 4*EIz/L

        k_elem += np.triu(k_elem, k=1).T
        return k_elem

    @property
    def mass_matrix_local(self):
        """Element mass matrix (as formulated in local system)"""

        # Ix: "Polar moment of inertia"
        rho = self.properties['rho']
        Ix = self.properties['Iy'] + self.properties['Iz']
        A = self.properties['A']

        rx2 = Ix/A
        L = self.length
        L2 = L**2

        m_elem = np.zeros((12, 12))

        m_elem[0, 0] = 140
        m_elem[0, 6] = 70

        m_elem[1, 1] = 156
        m_elem[1, 5] = 22*L
        m_elem[1, 7] = 54
        m_elem[1, 11] = -13*L

        m_elem[2, 2] = 156
        m_elem[2, 4] = -22*L
        m_elem[2, 8] = 54
        m_elem[2, 10] = 13*L

        m_elem[3, 3] = 140*rx2
        m_elem[3, 9] = 70*rx2

        m_elem[4, 4] = 4*L2
        m_elem[4, 8] = -13*L
        m_elem[4, 10] = -3*L2

        m_elem[5, 5] = 4*L2
        m_elem[5, 7] = 13*L
        m_elem[5, 11] = -3*L2

        m_elem[6, 6] = 140

        m_elem[7, 7] = 156
        m_elem[7, 11] = -22*L

        m_elem[8, 8] = 156
        m_elem[8, 10] = 22*L

        m_elem[9, 9] = 140*rx2

        m_elem[10, 10] = 4*L2

        m_elem[11, 11] = 4*L2

        m_elem += np.triu(m_elem, k=1).T
        m_elem *= (rho*A*L)/420
        return m_elem

    # @property
    # def point_mass_matrix_local(self):
    #     """Point mass matrix"""

    #     M_point = np.zeros((12, 12))

    #     m1 = self.point_properties['m1']
    #     m2 = self.point_properties['m2']

    #     # Exit early if possible...
    #     if m1 == 0 and m2 == 0:
    #         return M_point

    #     M_point[0:3, 0:3] = m1*np.identity(3)
    #     M_point[6:9, 6:9] = m2*np.identity(3)
    #     return M_point

    # @property
    # def mass_matrix_glob(self):
    #     """Element mass matrix (transformed to global system)"""

    #     # Mass matrix due to distributed element mass (density) and due to point masses
    #     m_elem = self.mass_matrix_local + self.point_mass_matrix_local

    #     m_glob = self.T.T @ m_elem @ self.T
    #     return m_glob

    @property
    def stiffness_matrix_glob(self):
        """Element stiffness matrix (transformed to global system)"""

        k_elem = self.stiffness_matrix_local
        k_glob = self.T.T @ k_elem @ self.T
        return k_glob

    @property
    def distributed_load_vector(self):
        """Distributed load vector (as formulated in local system)"""

        qx = self.distributed_loads['qx']
        qy = self.distributed_loads['qy']
        qz = self.distributed_loads['qz']
        mx = self.distributed_loads['mx']
        my = self.distributed_loads['my']
        mz = self.distributed_loads['mz']

        L = self.length
        L2 = L**2

        f_d_elem = np.empty((12, 1))

        f_d_elem[0] = qx*L/2
        f_d_elem[1] = qy*L/2 - mz
        f_d_elem[2] = qz*L/2 + my
        f_d_elem[3] = mx*L/2
        f_d_elem[4] = -qz*L2/12
        f_d_elem[5] = qy*L2/12

        f_d_elem[6] = qx*L/2
        f_d_elem[7] = qy*L/2 + mz
        f_d_elem[8] = qz*L/2 - my
        f_d_elem[9] = mx*L/2
        f_d_elem[10] = qz*L2/12
        f_d_elem[11] = -qy*L2/12

        return f_d_elem

    @property
    def distributed_load_vector_glob(self):
        """Distributed load vector (transformed to global system)"""

        self.T = self.transformation_matrix
        f_d_elem = self.distributed_load_vector

        # if self.distributed_loads_in_loc_system:
        #     f_d_elem = T.T @ f_d_elem

        return f_d_elem

    @property
    def concentrated_load_vector(self):
        """Concentrated load vector (here independent of coordinate system)"""

        f_c_elem = np.empty((12, 1))
        for i, load_type in enumerate(self.CONC_LOAD_TYPES):
            f_c_elem[i] = self.concentrated_loads[load_type]

        return f_c_elem

    @property
    def concentrated_load_vector_glob(self):
        """Concentrated load vector (transformed to global system)"""

        f_c_elem = self.concentrated_load_vector

        # if self.concentrated_loads_in_loc_system:
        #     f_c_elem = T.T @ f_c_elem

        return f_c_elem

    # @property
    # def load_vector_glob(self):
    #     """Element load vector (transformed to global system)"""

    #     f_d_elem = self.distributed_load_vector_glob
    #     f_c_elem = self.concentrated_load_vector_glob
    #     f_elem_glob = f_d_elem + f_c_elem
    #     return f_elem_glob

    def update_distributed_loads(self, loads, is_loc_system):
        """
        Add the distributed loads to the element

        Notes:
            * Acceptable dictionary keys see __class__.DIST_LOAD_TYPES

        Args:
            :loads: dict with distributed loads
            :is_loc_system: flag, which, if True, makes loads be treated
                according to local coordinate system
        """

        if is_loc_system not in (True, False):
            raise ValueError

        self.distributed_loads_in_loc_system = is_loc_system

        for load_type in loads.keys():
            self.distributed_loads[load_type] += loads.get(load_type, 0)

    def add_pointmass(self, pointmass, loc_node_number):
        """
        Add a point mass

        Args:
            :pointmass: mass in SI units
            :loc_node_number: node to which loads are to be applied (1, 2)
        """

        if loc_node_number not in (1, 2):
            raise ValueError

        mass = f'm{loc_node_number}'  # see POINT_PROPS
        self.point_properties[mass] += pointmass

    @property
    def mass(self):
        """
        Return the mass of the element
        """

        # "Cross section mass"
        rho = self.properties['rho']
        A = self.properties['A']
        L = self.length

        # Point masses
        m1 = self.point_properties[f'm1']
        m2 = self.point_properties[f'm2']

        return rho*A*L + m1 + m2


def get_local_system_from_up(x_elem, up):
    """
    Return the y- and z-axis of a Cartesian coordinate system (local element
    coordinate system) for a given x-axis and a defined "up-direction"

    Args:
        :x_elem: x-axis (of the element)
        :up: up-direction

    Returns:
        :y_axis: y-axis
        :z_axis: z-axis

    * Values are returned as a tuple
    """

    if abs(1 - abs(np.dot(x_elem, up))) <= 1e-10:
        logger.error("up-direction and local x-axis are parallel")
        raise ValueError("up-direction and local x-axis are parallel")

    z_elem = unit_vector(vector_rejection(up, x_elem))
    y_elem = unit_vector(np.cross(z_elem, x_elem))
    return (y_elem, z_elem)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------
# Copyright 2019-2020 Airinnova AB and the FramAT authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------

# Author: Aaron Dettmann

"""
Solving
"""

import numpy as np


def solve(m):
    static_load_analysis(m)


def static_load_analysis(m):
    """
    Perform a static load analysis

    Returns:
        :U: global vector of nodal displacements
        :F_react: reaction loads
    """

    mat = m.results.get('matrices')
    K = mat.get('K')
    B = mat.get('B')
    F = mat.get('F')

    ndof = K.shape[0]
    b = np.zeros((B.shape[0], 1))

    # ===== Assemble the system of equations =====
    # Number of linear constraints
    n_lr = B.shape[0]
    Z = np.zeros((n_lr, n_lr))

    A_system = np.block([
        [K, B.T],
        [B, Z]
    ])
    x_system = np.block([
        [F],  # + F_accel
        [b]
    ])
    solution = np.linalg.solve(A_system, x_system)

    U = solution[0:ndof]
    F_react = solution[ndof:]

    m.results.get('matrices').set('U', U)
    m.results.get('matrices').set('F_react', F_react)

    #############
    beam = m.results.get('beam')[0]
    beam.set(
        'deformation',
        {
            'ux': U[0::6, ].flatten(),
            'uy': U[1::6, ].flatten(),
            'uz': U[2::6, ].flatten(),
            'thx': U[3::6, ].flatten(),
            'thy': U[4::6, ].flatten(),
            'thz': U[5::6, ].flatten(),
        },
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------
# Copyright 2019-2020 Airinnova AB and the FramAT authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------

# Author: Aaron Dettmann

"""
Model definition
"""

from uuid import uuid4

from mframework import FeatureSpec, ModelSpec, SchemadictValidators
import numpy as np

from ._meshing import AbstractBeamMesh
from ._run import run_model
from ._util import Schemas as S

# Register custom 'schemadict' types
SchemadictValidators.register_type(AbstractBeamMesh)
SchemadictValidators.register_type(np.ndarray)

# TODO:
# - acceleration state (define on beam level, or 'global'?)


# =================
# ===== MODEL =====
# =================
mspec = ModelSpec()

tmp_doc = "The *{X}* feature is optional and allows to define sets of constant \
           {X} properties. When defining the properties for a specific beam \
           (or parts of it), you may refer to a {X} set using its UID. You may \
           define as many sets as you like."

# ===== Material =====
fspec = FeatureSpec()
fspec.add_prop_spec(
    'E',
    S.pos_number,
    required=True,
    doc="Young's modulus [N/m²]"
)
fspec.add_prop_spec(
    'G',
    S.pos_number,
    required=True,
    doc="Shear modulus [N/m²]"
)
fspec.add_prop_spec(
    'rho',
    S.pos_number,
    required=True,
    doc="Density [kg/m³]"
)
mspec.add_feature_spec(
    'material',
    fspec,
    singleton=False,
    required=False,
    doc=tmp_doc.format(X='material'),
    uid_required=True,
)

# ===== Cross-section =====
fspec = FeatureSpec()
fspec.add_prop_spec(
    'A',
    S.pos_number,
    required=True,
    doc="Area [m²]"
)
fspec.add_prop_spec(
    'Iy',
    S.pos_number,
    required=True,
    doc="Second moment of area about the local y-axis [m⁴]"
)
fspec.add_prop_spec(
    'Iz',
    S.pos_number,
    required=True,
    doc="Second moment of area about the local z-axis [m⁴]"
)
fspec.add_prop_spec(
    'J',
    S.pos_number,
    required=True,
    doc="Torsional constant [m⁴]"
)
mspec.add_feature_spec(
    'cross_section',
    fspec,
    singleton=False,
    required=False,
    doc=tmp_doc.format(X='cross section'),
    uid_required=True,
)

# ===== Beam =====
tmp_doc = "Define a constant {X} for a section of a beam. Refer to the start \
          of the beam section with the key 'from' followed by a node UID, \
          and refer to the end of the section with the key 'to'. "

fspec = FeatureSpec()
fspec.add_prop_spec(
    'node',
    S.vector3x1,
    singleton=False,
    doc="Add a named beam node, and defines its coordinates in a global \
         coordinate system. A beam requires at least two nodes. Note that you \
         must provide a UID.",
    uid_required=True,
)
fspec.add_prop_spec(
    'orientation',
    {
        '$required_keys': ['from', 'to', 'up'],
        'from': S.string,
        'to': S.string,
        'up': S.vector3x1
    },
    singleton=False,
    doc=tmp_doc.format(X='beam cross section orientation') +
        "The key 'up' is followed by a list (vector) indicating the direction \
        of the local z-axis of the beam element. The 'up' vector does not have \
        to be a unit vector."
)
fspec.add_prop_spec(
    'material',
    {
        '$required_keys': ['from', 'to', 'uid'],
        'from': S.string,
        'to': S.string,
        'uid': S.string
    },
    singleton=False,
    doc=tmp_doc.format(X='material') +
        "The key 'uid' must refer to a material UID defined in the 'material' feature.",
)
fspec.add_prop_spec(
    'cross_section',
    {'$required_keys': ['from', 'to', 'uid'], 'from': S.string, 'to': S.string, 'uid': S.string},
    singleton=False,
    doc=tmp_doc.format(X='cross section') +
    "The key 'uid' must refer to a cross section UID defined in the 'cross_section' feature.",
)
fspec.add_prop_spec(
    'point_load',
    {'$required_keys': ['at', 'load'], 'at': S.string, 'load': S.vector6x1},
    singleton=False,
    doc="Add a point load"
)
fspec.add_prop_spec(
    'nelem',
    S.pos_int,
    singleton=True,
    doc="Define the number of element for the beam object. The number will \
         apply to the whole polygonal chain. Note that the number is only \
         approximate, and the actual element number is determined by the \
         number and location of the named nodes."
)
mspec.add_feature_spec(
    'beam',
    fspec,
    singleton=False,
    required=False,
    doc="With the 'beam' feature you can add as many beams as needed for your \
         model. The beam geometry is defined with so-called 'named nodes'. \
         These are special nodes which have a UID and which together make up a \
         polygonal chain. In addition, you must also specify the cross section \
         orientation. Beam properties (material and cross-section data) has to \
         be defined for the entire beam length. Optionally, you can define \
         loads or mass properties for an individual beam."
)

# ===== Boundary conditions =====
fspec = FeatureSpec()
fspec.add_prop_spec(
    'fix',
    {
        '$required_keys': ['node', 'fix'],
        'node': S.string,
        'fix': {'type': list, 'min_len': 1, 'max_len': 6, 'item_types': str}
    },
    singleton=False,
    doc="Fix a beam node"
)
fspec.add_prop_spec(
    'connect',
    {
        '$required_keys': ['node1', 'node2', 'fix'], 'node1': S.string, 'node2': S.string,
        'fix': {'type': list, 'min_len': 1, 'max_len': 6, 'item_types': str}
    },
    singleton=False,
    doc="Connect two beam nodes"
)
mspec.add_feature_spec(
    'bc',
    fspec,
    singleton=True,
    required=True,
    doc="Boundary conditions"
)

# ===== Study =====
fspec = FeatureSpec()
fspec.add_prop_spec(
    'type',
    S.string,
    doc="Define a study type"
)
mspec.add_feature_spec('study', fspec, singleton=True, required=True, doc='Cross-section properties')

# ===== Post-proc =====
fspec = FeatureSpec()
fspec.add_prop_spec(
    'plot_settings',
    {
        'show': {'type', bool},
        'linewidth': S.pos_number,
        'markersize': S.pos_number,
        'fontsize': S.pos_int,
    },
    doc="General plot settings"
)
fspec.add_prop_spec(
    'plot',
    {
        'type': tuple,
        'allowed_items': (
            'deformed',
            'node_uids',
            'nodes',
            'undeformed',
        ),
    },
    singleton=False,
    doc="Add a geometry plot"
)
mspec.add_feature_spec(
    'post_proc',
    fspec,
    singleton=True,
    required=True,
    doc='Cross-section properties'
)

# ===================
# ===== RESULTS =====
# ===================
rspec = ModelSpec()

# ===== Mesh =====
fspec = FeatureSpec()
fspec.add_prop_spec(
    'named_nodes',
    {'type': dict},
    singleton=True,
    doc="Mapping of named nodes to global node numbers"
)
fspec.add_prop_spec(
    'nbeam',
    S.pos_int,
    singleton=True,
    doc=""
)
fspec.add_prop_spec(
    'nelem',
    S.pos_int,
    singleton=True,
    doc=""
)
fspec.add_prop_spec(
    'nnode',
    S.pos_int,
    singleton=True,
    doc=""
)
fspec.add_prop_spec(
    'ndof',
    S.pos_int,
    singleton=True,
    doc=""
)
fspec.add_prop_spec(
    'abm',
    {'type': AbstractBeamMesh},
    singleton=True,
    doc=""
)
rspec.add_feature_spec(
    'mesh',
    fspec,
    singleton=True,
    required=False,
    doc="Mesh"
)

# ===== Beam =====
fspec = FeatureSpec()
fspec.add_prop_spec(
    'named_node',
    S.string,
    singleton=False,
    doc="List of named nodes belonging to beam"
)
fspec.add_prop_spec(
    'mesh',
    {'type': AbstractBeamMesh},
    singleton=True,
    doc="List of named nodes belonging to beam"
)
fspec.add_prop_spec(
    'elements',
    {'type': list},
    singleton=True,
    doc="List of elements"
)
fspec.add_prop_spec(
    'deformation',
    {
        'ux': {'type': np.ndarray},
        'uy': {'type': np.ndarray},
        'uz': {'type': np.ndarray},
        'thx': {'type': np.ndarray},
        'thy': {'type': np.ndarray},
        'thz': {'type': np.ndarray},
    },
    singleton=True,
    doc="List of elements"
)
rspec.add_feature_spec(
    'beam',
    fspec,
    singleton=False,
    required=False,
    doc='Beam'
)

# ===== Deformation =====
fspec = FeatureSpec()
fspec.add_prop_spec('K', {'type': np.ndarray}, doc="TODO")
fspec.add_prop_spec('M', {'type': np.ndarray}, doc="TODO")
fspec.add_prop_spec('F', {'type': np.ndarray}, doc="TODO")
fspec.add_prop_spec('U', {'type': np.ndarray}, doc="TODO")
fspec.add_prop_spec('B', {'type': np.ndarray}, doc="TODO")
fspec.add_prop_spec('F_react', {'type': np.ndarray}, doc="TODO")
rspec.add_feature_spec(
    'matrices',
    fspec,
    singleton=True,
    doc='System matrices'
)

# ===== Deformation =====
fspec = FeatureSpec()
fspec.add_prop_spec(
    'max',
    S.pos_number,
    required=True,
    doc="Maximum deformation"
)
rspec.add_feature_spec(
    'deformation',
    fspec,
    doc='Deformation'
)

mspec.results = rspec


# ===== MODEL =====
class Model(mspec.user_class):
    def run(self):
        super().run()
        run_model(self)
        return self.results

    @classmethod
    def example(cls, example='cantilever'):
        if example == 'cantilever':
            return get_example_cantilever()
        elif example == 'helix':
            return get_example_helix()
        else:
            raise ValueError('unknown model: {example!r}')


def init_examples():
    model = Model()
    mat = model.add_feature('material', uid='dummy')
    mat.set('E', 1)
    mat.set('G', 1)
    mat.set('rho', 1)

    cs = model.add_feature('cross_section', uid='dummy')
    cs.set('A', 1)
    cs.set('Iy', 1)
    cs.set('Iz', 1)
    cs.set('J', 1)
    return model


def get_example_cantilever():
    model = init_examples()

    beam = model.add_feature('beam')
    beam.add('node', [0, 0, 0], uid='root')
    beam.add('node', [1, 0, 0], uid='tip')
    beam.set('nelem', 10)
    beam.add('material', {'from': 'root', 'to': 'tip', 'uid': 'dummy'})
    beam.add('cross_section', {'from': 'root', 'to': 'tip', 'uid': 'dummy'})
    beam.add('orientation', {'from': 'root', 'to': 'tip', 'up': [0, 0, 1]})
    beam.add('point_load', {'at': 'tip', 'load': [0, 0, -1, 0, 0, 0]})

    bc = model.set_feature('bc')
    bc.add('fix', {'node': 'root', 'fix': ['all']})
    return model


def get_example_helix():
    model = init_examples()

    model.get('material')[0].set('E', 1e5)
    model.get('material')[0].set('G', 1e4)

    def helix(a, b, c, t):
        """
        Make a Helix
        See: https://en.wikipedia.org/wiki/Helix
        """

        x = a*np.cos(t)
        y = b*np.sin(t)
        z = c*t
        return (x, y, z)

    # Generate node coordinates
    t = np.linspace(0, 20, num=200)
    x, y, z = helix(a=10, b=5, c=0.5, t=t)

    beam = model.add_feature('beam')
    for x_coord, y_coord, z_coord in zip(x, y, z):
        beam.add('node', [x_coord, y_coord, z_coord], uid=str(uuid4()))

    first_uid = beam.get_uid('node', 'first')
    last_uid = beam.get_uid('node', 'last')

    beam.set('nelem', 1)
    beam.add('material', {'from': first_uid, 'to': last_uid, 'uid': 'dummy'})
    beam.add('cross_section', {'from': first_uid, 'to': last_uid, 'uid': 'dummy'})
    beam.add('orientation', {'from': first_uid, 'to': last_uid, 'up': [0, 0, 1]})
    beam.add('point_load', {'at': last_uid, 'load': [0, 0, -1, 0, 0, 0]})

    bc = model.set_feature('bc')
    bc.add('fix', {'node': first_uid, 'fix': ['all']})
    return model



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------
# Copyright 2019-2020 Airinnova AB and the FramAT authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------

# Author: Aaron Dettmann

"""
Solving
"""

import itertools
from numbers import Number


def enumerate_with_step(iterable, start=0, step=1):
    """
    TODO
    https://stackoverflow.com/questions/24290025/python-enumerate-downwards-or-with-a-custom-step
    """

    for x in iterable:
        yield (start, x)
        start += step


def pairwise(iterable):
    """
    Return a new iterator which yields pairwise items

    s --> (s0,s1), (s1,s2), (s2, s3), ...

    See: https://docs.python.org/3/library/itertools.html#itertools-recipes
    """

    a, b = itertools.tee(iterable)
    next(b, None)

    return zip(a, b)


class Schemas:
    any_int = {'type': int}
    any_num = {'type': Number}
    pos_int = {'type': int, '>': 0}
    pos_number = {'type': Number, '>': 0}
    string = {'type': str, '>': 0}
    vector3x1 = {'type': list, 'min_len': 3, 'max_len': 3, 'item_types': Number}
    vector6x1 = {'type': list, 'min_len': 6, 'max_len': 6, 'item_types': Number}





#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------
# Copyright 2019-2020 Airinnova AB and the FramAT authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------

# Author: Aaron Dettmann

"""
Assembly
"""

import numpy as np

from ._util import enumerate_with_step
from ._element import Element
from ._log import logger


def create_system_matrices(m):
    """
    Assemble global tensors:

        * :K: global stiffness matrix
        * :M: global mass matrix
        * :F: global load vector
    """

    r = m.results
    abm = m.results.get('mesh').get('abm')

    # Create a stiffness matrix for each beam
    K_per_beam = []
    M_per_beam = []
    F_per_beam = []
    for i, (mbeam, rbeam) in enumerate(zip(m.iter('beam'), r.iter('beam'))):
        ndof_beam = Element.DOF_PER_NODE*(abm.nelem_beam(i) + 1)
        K_beam = np.zeros((ndof_beam, ndof_beam))
        M_beam = np.zeros((ndof_beam, ndof_beam))
        F_beam = np.zeros((ndof_beam, 1))

        for k, abel in enumerate_with_step(abm.beams[i].values(), step=6):
            phy_elem = Element.from_abstract_element(abel)
            K_beam[k:k+12, k:k+12] += phy_elem.stiffness_matrix_glob
            M_beam[k:k+12, k:k+12] += phy_elem.mass_matrix_glob
            F_beam[k:k+12] += phy_elem.load_vector_glob

        K_per_beam.append(K_beam)
        M_per_beam.append(M_beam)
        F_per_beam.append(F_beam)

    ndof_total = Element.DOF_PER_NODE*abm.nnodes
    K = np.zeros((ndof_total, ndof_total))
    M = np.zeros((ndof_total, ndof_total))
    F = np.zeros((ndof_total, 1))

    # Paste each beam tensor into the global tensor
    paste_range = [0, 0]
    for K_beam, M_beam, F_beam in zip(K_per_beam, M_per_beam, F_per_beam):
        rx = K_beam.shape[0]
        from_r = paste_range[1]
        to_r = from_r + rx
        paste_range = [from_r, to_r]
        K[from_r:to_r, from_r:to_r] = K_beam
        M[from_r:to_r, from_r:to_r] = M_beam
        F[from_r:to_r] += F_beam

    mat = r.set_feature('matrices')
    mat.set('K', K)
    mat.set('M', M)
    mat.set('F', F)


def create_bc_matrices(m):
    """Assemble the constraint matrix"""

    r = m.results
    abm = r.get('mesh').get('abm')
    ndof = Element.DOF_PER_NODE*(abm.nnodes)

    mbc = m.get('bc')

    B_tot = np.array([])

    # ----- Fix DOFs -----
    for fix in mbc.get('fix', []):
        num_node = abm.glob_nums[fix['node']]
        B = fix_dof(num_node, ndof, fix['fix'])
        B_tot = np.vstack((B_tot, B)) if B_tot.size else B

    # ----- Multipoint constraints (MPC) -----
    for con in mbc.get('connect', []):
        pass
        # TODO

    m.results.get('matrices').set('B', B_tot)


def fix_dof(node_number, total_ndof, dof_constraints):
    """
    Return part of constraint matrix B for fixed degrees of freedom

    Note:
        * Only non-zero rows are returned. If, say, three dof are fixed, then
          B will have size 3xndof

    Args:
        :node_number: node_number
        :total_ndof: total number of degrees of freedom
        :dof_constraints: list with dofs to be fixed
    """

    B = np.array([])
    pos_dict = {'ux': 0, 'uy': 1, 'uz': 2, 'tx': 3, 'ty': 4, 'tz': 5}

    for constraint in dof_constraints:
        if constraint == 'all':
            B = np.zeros((6, total_ndof))
            B[0:6, 6*node_number:6*node_number+6] = np.eye(6)
            break
        else:
            pos = pos_dict[constraint]

            B_row = np.zeros((1, total_ndof))
            B_row[0, 6*node_number+pos] = 1
            B = np.vstack((B, B_row)) if B.size else B_row

    return B


def connect(node1_number, node2_number, uid1, uid2, total_ndof, dof_constraints, frame):
    """
    Make a connection between two nodes

    Note:
        * The two nodes may belong to the same or to different beams

    Args:
        :node1_number: number of first node
        :node2_number: number of second node
        :uid1: UID of first node
        :uid2: UID of second node
        :total_ndof: total number of degrees of freedom
        :dof_constraints: list with dofs to be fixed (NOT YET IMPLEMENTED)
    """

    N1 = np.eye(6)
    N2 = -np.eye(6)

    x1 = frame.finder.nodes.by_uid[uid1].coord
    x2 = frame.finder.nodes.by_uid[uid2].coord
    x1 = np.asarray(x1)
    x2 = np.asarray(x2)
    dx, dy, dz = x1 - x2

    N2[0, 4] = -dz
    N2[0, 5] = dy
    N2[1, 3] = dz
    N2[1, 5] = -dx
    N2[2, 3] = -dy
    N2[2, 4] = dx

    B = np.zeros((6, frame.ndof))
    B[0:6, 6*node1_number:6*node1_number+6] = N1
    B[0:6, 6*node2_number:6*node2_number+6] = N2
    return B



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------
# Copyright 2019-2020 Airinnova AB and the FramAT authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------

# Author: Aaron Dettmann

"""
Module for creating the geometric mesh
"""

from collections import OrderedDict
from math import ceil

from mframework import FeatureSpec, UIDDict
import numpy as np

from ._element import Element
from ._log import logger
from ._util import Schemas as S
from ._util import pairwise

# TODO: Element --> Warning when overwriting data (first, from 'b' to 'c', then, from 'a' to 'c')
# TODO: need check that all material, cross section data is set from start to end....


def create_mesh(m):
    """Top-level function for mesh creation"""

    r = m.results
    abm = AbstractBeamMesh()

    for i, mbeam in enumerate(m.iter('beam')):
        logger.info(f"Meshing beam with index {i}...")

        # Support points (= named nodes)
        sup_points = OrderedDict()
        for uid, coord in mbeam.iter_uids('node'):
            sup_points[uid] = coord

        # Element lookup object for the new beam
        beam_id = abm.add_beam_line(sup_points, n=mbeam.get('nelem'))
        logger.info(f"Beam {i} has {abm.nelem_beam(beam_id)} elements")

        # ----- Beam element properties -----
        for pdef in mbeam.get('orientation'):
            for elem in abm.beams.iter_from_to(beam_id, pdef['from'], pdef['to']):
                elem.set('up', pdef['up'])

        for pdef in mbeam.get('material'):
            mat = m.get('material', uid=pdef['uid'])
            for elem in abm.beams.iter_from_to(beam_id, pdef['from'], pdef['to']):
                for p in ('E', 'G', 'rho'):
                    elem.set(p, mat.get(p))

        for pdef in mbeam.get('cross_section'):
            mat = m.get('cross_section', uid=pdef['uid'])
            for elem in abm.beams.iter_from_to(beam_id, pdef['from'], pdef['to']):
                for p in ('A', 'Iy', 'Iz', 'J'):
                    elem.set(p, mat.get(p))

        # ----- Loads -----
        for pdef in mbeam.get('point_load'):
            elem = abm.beams.get_by_uid(beam_id, pdef['at'])
            node_id = 1 if elem.p1.uid == pdef['at'] else 2
            elem.add('point_load', {'load': pdef['load'], 'node': node_id, 'local_sys': True})

        # for pdef in mbeam.get('distr_load'):
        #     for elem in abm.beams.iter_from_to(beam_id, pdef['from'], pdef['to']):
        #         ...

        # ----- Additional mass -----
        # for pdef in mbeam.get('point_mass'):
        #     elem = abm.beams.get_by_uid(beam_id, pdef['at'])
        #     node_id = 1 if elem.p1.uid == pdef['at'] else 2
        #     elem.add('point_load', {'load': pdef['load'], 'node': node_id, 'local_sys': True})

    logger.info(f"Abstract mesh created")
    logger.info(f"Discretisation:")
    logger.info(f"--> n_beams: {abm.nbeams:4d}")
    logger.info(f"--> n_elems: {abm.nelems:4d}")
    logger.info(f"--> n_nodes: {abm.nnodes:4d}")

    r.set_feature('mesh').set('abm', abm)


class Point:

    def __init__(self, coord, *, rel_coord=None, uid=None):
        assert uid is None or isinstance(uid, str)
        assert isinstance(coord, (list, np.ndarray)) and len(coord) == 3

        self.coord = np.array(coord)
        self.rel_coord = rel_coord
        self.uid = uid

    def __repr__(self):
        return f"{self.__class__.__qualname__}({self.coord!r}, {self.rel_coord!r}, {self.uid!r})"


class LineSegment:

    def __init__(self, p1, p2):
        assert isinstance(p1, Point) and isinstance(p2, Point)

        self.p1 = p1
        self.p2 = p2

        # Set the relative coordinates of the start and end point
        self.p1.rel_coord = 0
        self.p2.rel_coord = 1

        # Line direction
        self.dir = self.p2.coord - self.p1.coord
        self.len = np.linalg.norm(self.dir)
        self.all_points = [p1, p2]

    def __repr__(self):
        return f"{self.__class__.__qualname__}({self.p1!r}, {self.p2!r})"

    @property
    def from_uid(self):
        return self.p1.uid

    @property
    def to_uid(self):
        return self.p2.uid

    def split_into(self, n):
        """
        Split segment into smaller segments

        The new line segment will have n + 1 nodes.
        """

        assert isinstance(n, int) and n > 0

        self.all_points = [self.p1, ]

        for i in range(1, n):
            rel_coord = i/n
            self.all_points.append(
                Point(self.p1.coord + rel_coord*self.dir, rel_coord=rel_coord, uid=None)
            )

        self.all_points.append(self.p2)
        return self.all_points

    def iter_points(self, *, exclude_last=False):
        all_points = self.all_points[:-1] if exclude_last else self.all_points
        for p in all_points:
            yield p


class PolygonalChain:

    def __init__(self, sup_points, n=10):
        """
        Polygonal chain

        Args:
            :sup_points: (dict) key: support point uid, value: nodes coordinates
            :n: (int) number of elements for the entire chain

        Attr:
            :segments: (list) line segment
            :len: (float) length of the polygon
        """

        # Ensure that dictionary items are ordered correctly
        assert isinstance(sup_points, OrderedDict)
        assert isinstance(n, int)

        self.node_uids = sup_points.keys()
        self.segments = []
        self.len = 0

        # Add line segments
        points = []
        for uid, coord in sup_points.items():
            points.append(Point(coord, rel_coord=None, uid=uid))

        edges = []
        for p1, p2 in pairwise(points):
            edges.append(LineSegment(p1, p2))

        for seg in edges:
            self.add_segment(seg)

        self.set_node_num(n)

    def __repr__(self):
        return f"< {self.__class__.__qualname__} for {self.node_uids!r} >"

    def add_segment(self, seg):
        assert isinstance(seg, LineSegment)
        self.segments.append(seg)
        self.len += seg.len

    def set_node_num(self, n):
        for seg in self.segments:
            n_seg = ceil(n*(seg.len/self.len))
            seg.split_into(n_seg)

    def iter_points(self):
        curr_len = 0
        # All but last segments...
        for seg in self.segments[:-1]:
            for p in seg.iter_points(exclude_last=True):
                eta_poly = curr_len/self.len + p.rel_coord*(seg.len/self.len)
                yield Point(p.coord, rel_coord=eta_poly, uid=p.uid)
            curr_len += seg.len

        # Last segment...
        seg = self.segments[-1]
        for p in seg.iter_points(exclude_last=False):
            eta_poly = curr_len/self.len + p.rel_coord*(seg.len/self.len)  # TODO: duplicate
            yield Point(p.coord, rel_coord=eta_poly, uid=p.uid)


# ===== Abstract beam element =====

schema_load = {'load': S.vector6x1, 'node': S.pos_int, 'local_sys': {'type': bool}}
schema_mass = {'mass': {'type': S.pos_number}, 'node': S.pos_int}

fspec = FeatureSpec()

for p in Element.PROP_TYPES:
    fspec.add_prop_spec(p, S.pos_number)

fspec.add_prop_spec('up', S.vector3x1)
fspec.add_prop_spec('point_load', schema_load, singleton=False)
fspec.add_prop_spec('dist_load', schema_load, singleton=False)
fspec.add_prop_spec('point_mass', schema_mass, singleton=False)

# =================================


class AbstractEdgeElement(fspec.user_class):

    def __init__(self, p1, p2):
        """
        Edge element defined by spatial coordinates and abstract attributes

        Args:
            :p1: (obj) Point object of start point
            :p2: (obj) Point object of end point
        """

        super().__init__()

        assert isinstance(p1, Point) and isinstance(p2, Point)
        self.p1 = p1
        self.p2 = p2

    def __repr__(self):
        return f"< {self.__class__.__qualname__}({self.p1!r}, {self.p2!r}) >"


class AbstractBeamMesh:

    def __init__(self):
        """
        Mesh consisting of multiple, disconnected polygonal chains. The mesh
        may have globally defined abstract attributes.

        Attr:
            :beams: (list) list of beam look-up objects
        """

        # Each beam has a lookup dict object
        self.beams = UIDDict()

        # Bookkeeping
        self._num_beams = -1  # Number of beams -1
        self._num_elems = -1  # Number of elements -1
        self._num_nodes = -1  # Number of nodes -1

        # Global node numbers of named nodes
        self.glob_nums = {}

    def __repr__(self):
        return f"< {self.__class__.__qualname__} for {self.glob_nums!r} >"

    def add_beam_line(self, sup_points, n=10):
        """
        TODO
        """

        self._num_beams += 1
        self._num_nodes += 1

        polygon = PolygonalChain(sup_points, n=n)
        for (num1, p1), (num2, p2) in pairwise(enumerate(polygon.iter_points(), start=self._num_nodes)):
            self._num_elems += 1
            self.beams[self._num_beams] = AbstractEdgeElement(p1, p2)

            if p1.uid is not None:
                self.glob_nums[p1.uid] = num1
                try:
                    self.beams.assign_uid(self._num_beams, p1.uid)
                except KeyError:
                    pass
            if p2.uid is not None:
                self.glob_nums[p2.uid] = num2
                try:
                    self.beams.assign_uid(self._num_beams, p2.uid)
                except KeyError:
                    pass

        self._num_nodes += len(self.beams[self._num_beams])
        return self._num_beams

    @property
    def nbeams(self):
        return self._num_beams + 1

    def nelem_beam(self, beam_id):
        return len(self.beams[beam_id])

    @property
    def nelems(self):
        return self._num_elems + 1

    @property
    def nnodes(self):
        return sum(len(v.values()) + 1 for v in self.beams.values())

    def ndofs(self, ndof_per_node=6):
        return self.nnodes*ndof_per_node

    def get_sup_points(self, beam_id):
        xyz = self.beams[beam_id][0].p1.coord.reshape((1, 3))
        for elem in self.beams[beam_id].values():
            xyz = np.append(xyz, elem.p2.coord.reshape((1, 3)), axis=0)
        return xyz

    def get_lims_beam(self, beam_id):
        xyz = self.get_sup_points(beam_id)
        x_max = np.max(xyz[:, 0])
        y_max = np.max(xyz[:, 1])
        z_max = np.max(xyz[:, 2])
        x_min = np.min(xyz[:, 0])
        y_min = np.min(xyz[:, 1])
        z_min = np.min(xyz[:, 2])
        return (x_min, x_max), (y_min, y_max), (z_min, z_max)

    def get_lims(self):
        x_max = 0
        y_max = 0
        z_max = 0
        x_min = 0
        y_min = 0
        z_min = 0
        for beam_id in self.beams.keys():
            (x_minb, x_maxb), (y_minb, y_maxb), (z_minb, z_maxb) = self.get_lims_beam(beam_id)
            if x_maxb > x_max:
                x_max = x_maxb
            if y_maxb > y_max:
                y_max = y_maxb
            if z_maxb > z_max:
                z_max = z_maxb
            if x_minb < x_min:
                x_min = x_minb
            if y_minb < y_min:
                y_min = y_minb
            if z_minb < z_min:
                z_min = z_minb
        return (x_min, x_max), (y_min, y_max), (z_min, z_max)
